@@ -164,8 +164,10 @@ PRInt32 nsStandardURL::
 nsSegmentEncoder::EncodeSegmentCount(const char *str,
                                      const URLSegment &seg,
                                      PRInt16 mask,
-                                     nsAFlatCString &result)
+                                     nsAFlatCString &result,
+                                     PRBool &appended)
 {
+    appended = PR_FALSE;
     if (!str)
         return 0;
     PRInt32 len = 0;
@@ -197,11 +199,14 @@ nsSegmentEncoder::EncodeSegmentCount(const char *str,
         PRUint32 initLen = result.Length();
 
         // now perform any required escaping
-        if (NS_EscapeURL(str + pos, len, mask | escapeFlags, result))
+        if (NS_EscapeURL(str + pos, len, mask | escapeFlags, result)) {
             len = result.Length() - initLen;
+            appended = PR_TRUE;
+        }
         else if (str == encBuf.get()) {
             result += encBuf; // append only!!
             len = encBuf.Length();
+            appended = PR_TRUE;
         }
     }
     return len;
@@ -213,13 +218,11 @@ nsSegmentEncoder::EncodeSegment(const nsASingleFragmentCString &str,
                                 nsAFlatCString &result)
 {
     const char *text;
-    PRUint32 resultLen = result.Length();
-    EncodeSegmentCount(str.BeginReading(text), URLSegment(0, str.Length()), mask, result);
-    // since EncodeSegmentCount appends to result, we must check whether its length grew
-    if (result.Length() > resultLen)
+    PRBool encoded;
+    EncodeSegmentCount(str.BeginReading(text), URLSegment(0, str.Length()), mask, result, encoded);
+    if (encoded)
         return result;
-    else
-        return str;
+    return str;
 }
 
 PRBool nsStandardURL::
@@ -374,10 +377,10 @@ nsStandardURL::CoalescePath(netCoalesceFlags coalesceFlag, char *path)
 }
 
 PRUint32
-nsStandardURL::AppendSegmentToBuf(char *buf, PRUint32 i, const char *str, URLSegment &seg, const nsCString *escapedStr)
+nsStandardURL::AppendSegmentToBuf(char *buf, PRUint32 i, const char *str, URLSegment &seg, const nsCString *escapedStr, PRBool useEscaped)
 {
     if (seg.mLen > 0) {
-        if (escapedStr && !escapedStr->IsEmpty()) {
+        if (useEscaped) {
             seg.mLen = escapedStr->Length();
             memcpy(buf + i, escapedStr->get(), seg.mLen);
         }
@@ -409,15 +412,10 @@ nsStandardURL::BuildNormalizedSpec(const char *spec)
 
     // buffers for holding escaped url segments (these will remain empty unless
     // escaping is required).
-    nsCAutoString encUsername;
-    nsCAutoString encPassword;
-    nsCAutoString encHost;
-    nsCAutoString encDirectory;
-    nsCAutoString encBasename;
-    nsCAutoString encExtension;
-    nsCAutoString encParam;
-    nsCAutoString encQuery;
-    nsCAutoString encRef;
+    nsCAutoString encUsername, encPassword, encHost, encDirectory,
+      encBasename, encExtension, encParam, encQuery, encRef;
+    PRBool useEncUsername, useEncPassword, useEncHost, useEncDirectory,
+      useEncBasename, useEncExtension, useEncParam, useEncQuery, useEncRef;
 
     //
     // escape each URL segment, if necessary, and calculate approximate normalized
@@ -434,14 +432,14 @@ nsStandardURL::BuildNormalizedSpec(const char *spec)
     // appropriate encoding.
     {
         GET_SEGMENT_ENCODER(encoder);
-        approxLen += encoder.EncodeSegmentCount(spec, mUsername,  esc_Username,      encUsername);
-        approxLen += encoder.EncodeSegmentCount(spec, mPassword,  esc_Password,      encPassword);
-        approxLen += encoder.EncodeSegmentCount(spec, mDirectory, esc_Directory,     encDirectory);
-        approxLen += encoder.EncodeSegmentCount(spec, mBasename,  esc_FileBaseName,  encBasename);
-        approxLen += encoder.EncodeSegmentCount(spec, mExtension, esc_FileExtension, encExtension);
-        approxLen += encoder.EncodeSegmentCount(spec, mParam,     esc_Param,         encParam);
-        approxLen += encoder.EncodeSegmentCount(spec, mQuery,     esc_Query,         encQuery);
-        approxLen += encoder.EncodeSegmentCount(spec, mRef,       esc_Ref,           encRef);
+        approxLen += encoder.EncodeSegmentCount(spec, mUsername,  esc_Username,      encUsername,  useEncUsername);
+        approxLen += encoder.EncodeSegmentCount(spec, mPassword,  esc_Password,      encPassword,  useEncPassword);
+        approxLen += encoder.EncodeSegmentCount(spec, mDirectory, esc_Directory,     encDirectory, useEncDirectory);
+        approxLen += encoder.EncodeSegmentCount(spec, mBasename,  esc_FileBaseName,  encBasename,  useEncBasename);
+        approxLen += encoder.EncodeSegmentCount(spec, mExtension, esc_FileExtension, encExtension, useEncExtension);
+        approxLen += encoder.EncodeSegmentCount(spec, mParam,     esc_Param,         encParam,     useEncParam);
+        approxLen += encoder.EncodeSegmentCount(spec, mQuery,     esc_Query,         encQuery,     useEncQuery);
+        approxLen += encoder.EncodeSegmentCount(spec, mRef,       esc_Ref,           encRef,       useEncRef);
     }
 
     // do not escape the hostname, if IPv6 address literal, mHost will
@@ -449,6 +447,7 @@ nsStandardURL::BuildNormalizedSpec(const char *spec)
     // However, perform Unicode normalization on it, as IDN does.
     mHostEncoding = eEncoding_ASCII;
     if (mHost.mLen > 0) {
+        useEncHost = PR_FALSE;
         const nsCSubstring& tempHost =
             Substring(spec + mHost.mPos, spec + mHost.mPos + mHost.mLen);
         if (IsASCII(tempHost))
@@ -456,8 +455,10 @@ nsStandardURL::BuildNormalizedSpec(const char *spec)
         else {
             mHostEncoding = eEncoding_UTF8;
             if (gIDNService &&
-                NS_SUCCEEDED(gIDNService->Normalize(tempHost, encHost)))
+                NS_SUCCEEDED(gIDNService->Normalize(tempHost, encHost))) {
                 approxLen += encHost.Length();
+                useEncHost = PR_TRUE;
+            }
             else {
                 encHost.Truncate();
                 approxLen += mHost.mLen;
@@ -484,15 +485,15 @@ nsStandardURL::BuildNormalizedSpec(const char *spec)
 
     // append authority
     if (mUsername.mLen > 0) {
-        i = AppendSegmentToBuf(buf, i, spec, mUsername, &encUsername);
+        i = AppendSegmentToBuf(buf, i, spec, mUsername, &encUsername, useEncUsername);
         if (mPassword.mLen >= 0) {
             buf[i++] = ':';
-            i = AppendSegmentToBuf(buf, i, spec, mPassword, &encPassword);
+            i = AppendSegmentToBuf(buf, i, spec, mPassword, &encPassword, useEncPassword);
         }
         buf[i++] = '@';
     }
     if (mHost.mLen > 0) {
-        i = AppendSegmentToBuf(buf, i, spec, mHost, &encHost);
+        i = AppendSegmentToBuf(buf, i, spec, mHost, &encHost, useEncHost);
         net_ToLowerCase(buf + mHost.mPos, mHost.mLen);
         if (mPort != -1 && mPort != mDefaultPort) {
             nsCAutoString portbuf;
@@ -526,7 +527,7 @@ nsStandardURL::BuildNormalizedSpec(const char *spec)
         // record corrected (file)path starting position
         mPath.mPos = mFilepath.mPos = i - leadingSlash;
 
-        i = AppendSegmentToBuf(buf, i, spec, mDirectory, &encDirectory);
+        i = AppendSegmentToBuf(buf, i, spec, mDirectory, &encDirectory, useEncDirectory);
 
         // the directory must end with a '/'
         if (buf[i-1] != '/') {
@@ -534,7 +535,7 @@ nsStandardURL::BuildNormalizedSpec(const char *spec)
             mDirectory.mLen++;
         }
 
-        i = AppendSegmentToBuf(buf, i, spec, mBasename, &encBasename);
+        i = AppendSegmentToBuf(buf, i, spec, mBasename, &encBasename, useEncBasename);
 
         // make corrections to directory segment if leadingSlash
         if (leadingSlash) {
@@ -547,22 +548,22 @@ nsStandardURL::BuildNormalizedSpec(const char *spec)
 
         if (mExtension.mLen >= 0) {
             buf[i++] = '.';
-            i = AppendSegmentToBuf(buf, i, spec, mExtension, &encExtension);
+            i = AppendSegmentToBuf(buf, i, spec, mExtension, &encExtension, useEncExtension);
         }
         // calculate corrected filepath length
         mFilepath.mLen = i - mFilepath.mPos;
 
         if (mParam.mLen >= 0) {
             buf[i++] = ';';
-            i = AppendSegmentToBuf(buf, i, spec, mParam, &encParam);
+            i = AppendSegmentToBuf(buf, i, spec, mParam, &encParam, useEncParam);
         }
         if (mQuery.mLen >= 0) {
             buf[i++] = '?';
-            i = AppendSegmentToBuf(buf, i, spec, mQuery, &encQuery);
+            i = AppendSegmentToBuf(buf, i, spec, mQuery, &encQuery, useEncQuery);
         }
         if (mRef.mLen >= 0) {
             buf[i++] = '#';
-            i = AppendSegmentToBuf(buf, i, spec, mRef, &encRef);
+            i = AppendSegmentToBuf(buf, i, spec, mRef, &encRef, useEncRef);
         }
         // calculate corrected path length
         mPath.mLen = i - mPath.mPos;
@@ -1134,18 +1135,20 @@ nsStandardURL::SetUserPass(const nsACString &input)
     nsCAutoString buf;
     if (usernameLen > 0) {
         GET_SEGMENT_ENCODER(encoder);
+        PRBool ignoredOut;
         usernameLen = encoder.EncodeSegmentCount(userpass.get(),
                                                  URLSegment(usernamePos,
                                                             usernameLen),
                                                  esc_Username | esc_AlwaysCopy,
-                                                 buf);
+                                                 buf, ignoredOut);
         if (passwordLen >= 0) {
             buf.Append(':');
             passwordLen = encoder.EncodeSegmentCount(userpass.get(),
                                                      URLSegment(passwordPos,
                                                                 passwordLen),
                                                      esc_Password |
-                                                     esc_AlwaysCopy, buf);
+                                                     esc_AlwaysCopy, buf,
+                                                     ignoredOut);
         }
         if (mUsername.mLen < 0)
             buf.Append('@');
@@ -2028,9 +2031,11 @@ nsStandardURL::SetQuery(const nsACString &input)
 
     // encode query if necessary
     nsCAutoString buf;
+    PRBool encoded;
     GET_SEGMENT_ENCODER(encoder);
-    encoder.EncodeSegmentCount(query, URLSegment(0, queryLen), esc_Query, buf);
-    if (!buf.IsEmpty()) {
+    encoder.EncodeSegmentCount(query, URLSegment(0, queryLen), esc_Query,
+                               buf, encoded);
+    if (encoded) {
         query = buf.get();
         queryLen = buf.Length();
     }
@@ -2086,9 +2091,11 @@ nsStandardURL::SetRef(const nsACString &input)
 
     // encode ref if necessary
     nsCAutoString buf;
+    PRBool encoded;
     GET_SEGMENT_ENCODER(encoder);
-    encoder.EncodeSegmentCount(ref, URLSegment(0, refLen), esc_Ref, buf);
-    if (!buf.IsEmpty()) {
+    encoder.EncodeSegmentCount(ref, URLSegment(0, refLen), esc_Ref,
+                               buf, encoded);
+    if (encoded) {
         ref = buf.get();
         refLen = buf.Length();
     }
@@ -2156,17 +2163,20 @@ nsStandardURL::SetFileName(const nsACString &input)
         }
         else {
             nsCAutoString newFilename;
+            PRBool ignoredOut;
             GET_SEGMENT_ENCODER(encoder);
             basename.mLen = encoder.EncodeSegmentCount(filename, basename,
                                                        esc_FileBaseName |
                                                        esc_AlwaysCopy,
-                                                       newFilename);
+                                                       newFilename,
+                                                       ignoredOut);
             if (extension.mLen >= 0) {
                 newFilename.Append('.');
                 extension.mLen = encoder.EncodeSegmentCount(filename, extension,
                                                             esc_FileExtension |
                                                             esc_AlwaysCopy,
-                                                            newFilename);
+                                                            newFilename,
+                                                            ignoredOut);
             }
 
             if (mBasename.mLen < 0) {
